from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..ket import dump as _dump
from .quant import quant
from typing import Optional
from random import choices
from functools import reduce
from operator import add

class dump(_dump):
    """Dump qubits

    Create a dump with the state current state of the :class:`~ket.types.quant`
    ``q``.

    Creating a dump does not trigger quantum execution, but gathering any
    information from it does.

    :Example:

    .. code-block:: ket

        a = H(quant())
        b = ctrl(a, X, quant())
        d1 = dump(a|b)
        Y(a)
        d2 = dump(a|b)

        print(d1.show())
        print(d2.show())

    :param q: Qubits to dump.
    """
    
    def __init__(self, q : quant):
        q = reduce(add, q)
        super().__init__(q)
        self.size = len(q)
        padding = self.size%64
        if padding != 0:
            padding = 64-padding
        self._bit_size = self.size+padding
        self._states = None
    
    @property
    def states(self) -> list[int]:
        """List of basis states."""

        if self._states is None:
            self._states = [int(reduce(add, [f'{i:064b}' for i in reversed(state)]) , 2) for state in super().get_states()]

        return self._states

    def _convert_state(self, state : int) -> list[int]:
        state_bin = f'{state:0b}'
        padding = self._bit_size - len(state_bin)
        if padding != 0:
            state_bin = '0'*padding+state_bin    
        
        return [int(state_bin[i:i+64], 2) for i in reversed(range(0, self._bit_size, 64))]

    def amplitude(self, state : int) -> complex | list[complex]:
        """Return the amplitude of a given state
        
        If the state is entangled with a not dumped qubit, return a list of
        amplitudes.
        
        Args:
            state: Basis state.
        """

        ret = super().amplitude(self._convert_state(state))
        return ret[0] if len(ret) == 1 else ret
    
    def probability(self, state : int) -> float:
        """Return the measure probabilit of a basis state.
          
        Args:
            state: Basis state.
        """
        return super().probability(self._convert_state(state))

    def show(self, format : Optional[str] = None) -> str:
        r"""Return the quantum state as a string
        
        Use the format starting to change the print format of the basis states:

        * ``i``: print the state in the decimal base
        * ``b``: print the state in the binary base (default)
        * ``i|b<n>``: separate the ``n`` first qubits, the remaining print in the binary base
        * ``i|b<n1>:i|b<n2>[:i|b<n3>...]``: separate the ``n1, n2, n3, ...`` first qubits
        
        :Example:
            
        .. code-block:: ket

            q = quant(20)
            X(ctrl(H(q[0]), X, q[1:])[1::2])
            d = dump(q)

            print(d.show('i'))
            #|174762⟩		(50%)
            #0.707107                 ≅  1/√2
            #
            #|873813⟩		(50%)
            #0.707107                 ≅  1/√2
            print(d.show('b'))
            #|00101010101010101010⟩		(50%)
            #0.707107                 ≅  1/√2
            #
            #|11010101010101010101⟩		(50%)
            #0.707107                 ≅  1/√2
            print(d.show('i4'))
            #|2⟩|1010101010101010⟩		(50%)
            #0.707107                 ≅  1/√2
            #
            #|13⟩|0101010101010101⟩		(50%)
            #0.707107                 ≅  1/√2
            print(d.show('b5:i4'))
            #|00101⟩|2⟩|101010101010⟩		(50%)
            #0.707107                 ≅  1/√2
            #
            #|11010⟩|5⟩|010101010101⟩		(50%)
            #0.707107                 ≅  1/√2

        Args:
            format: Format string that matchs ``(i|b)\d*(:(i|b)\d+)*``.
        """
        
        if format == 'i' or format == 'b':
            format += str(self.size)

        return super().show(format if format is not None else "")

    __repr__ = lambda self : f"<Ket type 'dump'; {super().__repr__()}>"

class dump_measure():

    def __init__(self, q : quant):
        self.dump = dump(q)
        self.states = None
        self.probabilities = None
        self.sorted = False

    def get(self, n : Optional[int] = None, no_repeat : bool = False):

        if self.states is None:
            self.states = self.dump.get_states()
            self.probabilities = [self.dump.probability(i) for i in self.states]
        
        if n is None:
            return choices(self.states, weights=self.probabilities)[0]
        elif no_repeat:
            if not self.sorted:
                self.sorted = True
                sorted_states = [(s, p) for s, p in sorted(zip(self.states, self.probabilities), key = lambda x : x[1], reverse=True)]
                self.states = [s for s, _ in sorted_states]    
                self.probabilities = [p for _, p in sorted_states]
            
            return self.states[:n if n <= len(self.states) else None]
        else:
            return choices(self.states, weights=self.probabilities, k=n)
 


#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..ket import future, measure

future.__doc__ = \
    """Quantum computer's classical value

    Store a reference to a 64-bits integer available in the quantum computer. 

    The integer value are available to the classical computer only after the
    quantum execution.
    
    The follwing binary operations are available between
    :class:`~ket.types.future` variables and ``int``: 
        
        ``==``, ``!=``, ``<``, ``<=``,
        ``>``, ``>=``, ``+``, ``-``, ``*``, ``/``, ``<<``, ``>>``, ``and``,
        ``xor``, and ``or``.

    A :class:`~ket.types.future` variable is created with a quantum
    :func:`~ket.standard.measure` (1) , binary operation with a
    :class:`~ket.types.future` (2), or directly initialization with a ``int``
    (2).

    .. code-block:: ket

        q = H(quant(2))
        a = measure(q) # 1
        b = a*3        # 2
        c = future(42) # 3
        
    To pass an assignment expression to the quantum computer use the syntax:
    
    .. code-block:: ket

        <var>.set = <exp>

    :Example:

    .. code-block:: ket

        q = quant(2)
        with quant() as aux:
            done = future(False)
            while done != True:
                H(q)
                ctrl(q, X, aux)
                res = measure(aux)
                if res == 0:
                    done.set = True
                else:
                    X(q+aux)
            aux.free() 

    :type value: ``int``
    :param value: Classical integer to pass to the quantum computer.
    """

future.get.__doc__ = \
    """ Get the value from the quantum computer

    Note: 
        This method triggers the quantum execution.

    :Example:
    
    .. code-block:: ket

        q = H(quant(60))
        m = measure(q)
        print('result:', m.get())

    :rtype: ``int``
    """

future.set.__doc__ = \
    """Set the value

    :type other: :class:`~ket.types.future`
    :param other: New value.
    """ 
    
future.__radd__     = lambda self, other : (future(other) if not isinstance(other, future) else other) + self
future.__rsub__     = lambda self, other : (future(other) if not isinstance(other, future) else other) - self
future.__rmul__     = lambda self, other : (future(other) if not isinstance(other, future) else other) * self
future.__rtruediv__ = lambda self, other : (future(other) if not isinstance(other, future) else other) // self
future.__rdiv__     = future.__rtruediv__
future.__rlshift__  = lambda self, other : (future(other) if not isinstance(other, future) else other) << self
future.__rrshift__  = lambda self, other : (future(other) if not isinstance(other, future) else other) >> self
future.__rand__     = lambda self, other : (future(other) if not isinstance(other, future) else other) & self
future.__rxor__     = lambda self, other : (future(other) if not isinstance(other, future) else other) ^ self
future.__ror__      = lambda self, other : (future(other) if not isinstance(other, future) else other) | self

def _future_setattr(self, name, value):
    if name == 'set':
        self.set(value if isinstance(value, future) else future(value))
    else:
        self.__dict__[name] = value

future.__setattr__ = _future_setattr
future.__int__ = lambda self : self.get()
future.__repr__ = lambda self : f"<Ket type 'future'; {self.this.__repr__()}>"



#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..ket import report, metrics, context, config, build_info as ket_version

__all__ = ['ket_config', 'report', 'context', 'ket_version']

def ket_config(**params):
    for param in params:
        config(param, str(params[param]))

report.__doc__ = \
    """Return the current KQASM metrics 

    **Example:**

    .. code-block:: ket

        >>> q = quant(20)
        >>> ctrl(H(q[0]), X, q[1:])
        >>> metrics = report()
        >>> print(metrics)
        Qubits used:         	20
        Free qubits:         	0
        Allocated Qubits:    	20
          ↳ Max concurrently:	20
        Measurements:        	0
        Quantum gates:       	20
          ↳ X gate:         	19
          ↳ H gate:         	1
        Quantum CTRL gates:  	19
          ↳ 1 control:      	19
        Ket Bitwise plugins: 	0

    :rtype: :class:`ket.util.metrics`
    """

metrics.__repr__ = lambda self : '<Ket metrics; '+self.this.__repr__()+'>'


#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..ket import measure as _measure, exec_quantum, process_begin, process_end
from ..types import quant, future
from ..preprocessor import _ket_if, _ket_next
from .ctrl import *
from .adj import *

__all__ = ['run', 'inverse', 'control', 'ctrl', 'adj', 'around', 'measure', 'exec_quantum']

exec_quantum.__doc__ = \
    """Trigger the quantum execution

    Execute the top process and fullfil the :class:`~ket.types.future` and
    :class:`~ket.types.dump` variables.
    """

def measure(q : quant, free : bool = False) -> future:
    """Quantum measurement

    Measure the qubits of a :class:`~ket.types.quant` and return a
    :class:`~ket.types.future`.

    Args:
        q: Qubits to measure.
        free: If ``True``, free the qubits after the measuremet.
    """

    ret = _measure(q)
    if free:
        for i in q:
            end = _ket_if(_measure(i))
            X(i)
            _ket_next(end) 
        q.free()
    return ret

class run:
    """Execute in a new process
    
    Run the quantum operations in a new separated process.
    
    :Usage:

    .. code-block:: ket

        with run():
            ...

    """
    def __enter__ (self):
        process_begin()
    
    def __exit__ (self, type, value, tb):
        process_end()



from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..preprocessor import ketpp
from inspect import getsource
from ast import fix_missing_locations, parse
from os import path, PathLike
from types import ModuleType
from typing import Callable

__all__ = ['import_ket', 'from_import_ket']

def _import_ket(source : PathLike):
    with open(source, 'r') as file:
        tree = parse(file.read(), source)
    
    pp = ketpp()

    pp.visit(tree)
    fix_missing_locations(tree)

    return compile(tree, source, 'exec', optimize=2)
    
def _import_globals_ket(source : PathLike, globals):
    exec(_import_ket(source), globals)    

buildins = \
"""
from ket import *
from ket.ket import label, branch, jump
from ket.preprocessor import *
"""
 
def import_ket(source : PathLike) -> ModuleType:
    """Import Ket file.

    Import a Ket source file as a Python module.

    :param source: Ket source file.  
    """

    _, file_name = path.split(source)
    module = ModuleType(file_name[:-4])

    obj = compile(buildins, '<ket build-in functions>', 'exec', optimize=2)
    exec(obj, module.__dict__)

    _import_globals_ket(source, module.__dict__)
    return module

def from_import_ket(source : PathLike, *names : list[str]) -> tuple:
    """Import names from Ket file.

    Args:
        source: Ket source file.  
        names: Names to import from ``source``.
    """
    
    module = import_ket(source)
    return tuple(module.__dict__[name] for name in names)

def code_ket(func : Callable) -> Callable:
    """Parse as Ket function.
    
    Decorator to parse a Python function as a Ket Function.
    
    Worning:
        Do not use this decorator in a .ket file.    

    :Usage:

    .. code-block:: python
        
        @code_ket
        def func(a : quant, b : quant, c : quant):
            m0 = measure(a)
            m1 = measure(b)
            if m1 == 1: # need @code_ket to properly execute
                x(c)
            if m0 == 1: # need @code_ket to properly execute
                z(c)
    """

    if '__in_ket__' in globals() and globals()['__in_ket__']:
        return func

    in_ket = "__in_ket__{}__".format(func.__name__)
    if in_ket in func.__globals__ and func.__globals__[in_ket]: 
        return func
        
    doc = func.__doc__
    annotations = func.__annotations__

    buildins_more = buildins+'\n'+in_ket+' = True\n'
    buildins_obj = compile(buildins_more, '<ket build-in functions>', 'exec', optimize=2)
    exec(buildins_obj, func.__globals__)
    
    tree = parse(getsource(func), '<function ' + func.__name__ + '>')
    pp = ketpp()

    pp.visit(tree)
    fix_missing_locations(tree)

    obj =  compile(tree, '<code_ket function ' + func.__name__ + '>', 'exec', optimize=2)
    exec(obj, func.__globals__)
    func.__globals__[func.__name__].__doc__ = doc
    func.__globals__[func.__name__].__annotations__ = annotations

    return func.__globals__[func.__name__]  



#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import ast, copy

class ketpp (ast.NodeTransformer):

    def __init__(self):
        self.id_count = 0
        self.while_begin = None
        self.while_end = None
        self.in_while = False

    def visit_If(self, node):
        
        if_id = self.id_count
        self.id_count += 1
        
        self.generic_visit(node)
        node_cp = copy.deepcopy(node)
        
        _test_name = '_ket_temp_if_test'+str(if_id)
        test_name = ast.Name(id=_test_name, ctx=ast.Load())

        end_name = '_ket_tmp_if_end'+str(if_id)
        else_name = None
        if node.orelse:
            else_name = '_ket_tmp_if_else'+str(if_id)
            if_begin = ast.Assign(
                targets=[
                    ast.Tuple(
                        elts=[
                            ast.Name(id=else_name, ctx=ast.Store()),
                            ast.Name(id=end_name, ctx=ast.Store())
                        ],
                        ctx=ast.Store()
                    )
                ],
                value=ast.Call(
                    func=ast.Name(id = '_ket_if_else', ctx=ast.Load()),
                    args=[test_name],
                    keywords=[]
                )
            )
        else:
            if_begin = ast.Assign(
                targets=[
                    ast.Name(id=end_name, ctx=ast.Store()),
                ],
                value=ast.Call(
                    func=ast.Name(id = '_ket_if', ctx=ast.Load()),
                    args=[test_name],
                    keywords=[]
                )
            )

        if_body = [if_begin, *node.body]
        
        if else_name is not None:
            else_call = ast.Expr(value=ast.Call(
                func=ast.Name(id = '_ket_else', ctx=ast.Load()),
                args=[ast.Name(id=name, ctx=ast.Load()) for name in [else_name, end_name]],
                keywords=[]
            ))
            if_body.extend([else_call, *node.orelse])

        if_body.append(
            ast.Expr(value=ast.Call(
                func=ast.Name(id = '_ket_next', ctx=ast.Load()),
                args=[ast.Name(id=end_name, ctx=ast.Load())],
                keywords=[]
            ))
        )

        test_assing = ast.Assign(
            targets=[ast.Name(id=_test_name, ctx=ast.Store())], 
            value=node.test
        )

        type_check = ast.Call(
            func=ast.Name(id='_ket_is_future', ctx=ast.Load()),
            args=[test_name],
            keywords=[]
        )
 
        node_cp.test = test_name

        if_future = ast.If(
            test=type_check,
            body=if_body,
            orelse=[node_cp]
        )
        
        return [test_assing, if_future]
            
    def visit_While(self, node):
        while_begin_save = self.while_begin
        while_end_save = self.while_end

        while_id = self.id_count
        self.id_count += 1
        
        begin_name = '_ket_tmp_while_begin'+str(while_id)
        loop_name  = '_ket_tmp_while_loop'+str(while_id)
        end_name  = '_ket_tmp_while_end'+str(while_id)
        else_name = None
        self.while_end = end_name
        if node.orelse:
            else_name = '_ket_tmp_while_else'+str(while_id)
            while_begin = ast.Assign(
                targets=[
                    ast.Tuple(
                        elts=[ast.Name(id=name, ctx=ast.Store()) for name in [begin_name, loop_name, else_name, end_name]],
                        ctx=ast.Store()
                    )
                ],
                value=ast.Call(
                    func=ast.Name(id = '_ket_while_else', ctx=ast.Load()),
                    args=[],
                    keywords=[]
                )
            )
        else:
            while_begin = ast.Assign(
                targets=[
                    ast.Tuple(
                        elts=[ast.Name(id=name, ctx=ast.Store()) for name in [begin_name, loop_name, end_name]],
                        ctx=ast.Store()
                    )
                ],
                value=ast.Call(
                    func=ast.Name(id = '_ket_while', ctx=ast.Load()),
                    args=[],
                    keywords=[]
                )
            )
        
        test_name = '_ket_while_test'+str(while_id)
        test_assing = ast.Assign(
            targets=[ast.Name(id=test_name, ctx=ast.Store())], 
            value=node.test
        )

        type_check = ast.Call(
            func=ast.Name(id='_ket_is_future', ctx=ast.Load()),
            args=[ast.Name(id=test_name, ctx=ast.Load())],
            keywords=[]
        )
        
        body_call = ast.Expr(value=ast.Call(
            func=ast.Name(id= '_ket_while_body', ctx=ast.Load()),
            args=[
                ast.Name(id=test_name, ctx=ast.Load()),
                ast.Name(id=loop_name, ctx=ast.Load()),
                ast.Name(id=else_name if else_name is not None else end_name, ctx=ast.Load())
            ],
            keywords=[]
        )) 

        loop_call = ast.Expr(value=ast.Call(
            func=ast.Name(id= '_ket_loop', ctx=ast.Load()),
            args=[
                ast.Name(id=begin_name, ctx=ast.Load()),
                ast.Name(id=else_name if else_name is not None else end_name, ctx=ast.Load())
            ],
            keywords=[]
        ))
        
        else_call = [ast.Expr(value=ast.Call(
            func=ast.Name(id = '_ket_next', ctx=ast.Load()),
            args=[ast.Name(id=end_name, ctx=ast.Load())],
            keywords=[]
        ))] if else_name is not None else []
        
        node_cp = copy.deepcopy(node)
        in_while_save = self.in_while
        self.in_while = True
        self.generic_visit(node)
        self.in_while = in_while_save
        
        self.generic_visit(node_cp)
        
        ndone_name = '_ket_tmp_while_ndone'+str(while_id)

        if_not_future = ast.If(
            test=ast.Name(id=test_name, ctx=ast.Load()),
            body=[
                ast.Assign(
                    targets=[ast.Name(id=ndone_name, ctx=ast.Store())],
                    value=ast.Constant(value=True)
                ),
                ast.While(
                    test=ast.Name(id=ndone_name, ctx=ast.Load()),
                    body = [*node_cp.body, 
                        ast.If(
                            test=ast.UnaryOp(op=ast.Not(), operand=node_cp.test),
                            body=[
                                ast.Assign(
                                    targets=[ast.Name(id=ndone_name, ctx=ast.Store())],
                                    value=ast.Constant(value=False)
                                )
                            ],
                            orelse=[]
                        )
                    ],
                    orelse=[]
                ),
                ast.If(
                    test=ast.UnaryOp(
                        op=ast.Not(), 
                        operand=ast.Name(id=ndone_name, ctx=ast.Load())
                    ),
                    body=node_cp.orelse if node_cp.orelse else [ast.Pass()],
                    orelse=[]
                )
            ],
            orelse=node_cp.orelse
        )

        if_future = ast.If(
            test=type_check,
            body=[body_call, *node.body, loop_call, *else_call],
            orelse=[if_not_future]
        )

        self.while_begin = while_begin_save 
        self.while_end   = while_end_save   

        return [while_begin, test_assing, if_future]
    

    def visit_Break(self, node):
        return ast.Expr(value=ast.Call(
            func=ast.Name(id='_ket_goto', ctx=ast.Load()),
            args=[ast.Name(id=self.while_end, ctx=ast.Load())],
            keywords=[]
        )) if self.in_while else node

    def visit_Continue(self, node):
        return ast.Expr(value=ast.Call(
            func=ast.Name(id='_ket_goto', ctx=ast.Load()),
            args=[ast.Name(id=self.while_begin, ctx=ast.Load())],
            keywords=[]
        )) if self.in_while else node



#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .util import *
from .gates import *
from .types import *
from .standard import *
from .import_ket import *
from .util import __all__ as all_util
from .gates import __all__ as all_gate
from .types import __all__ as all_types
from .import_ket import __all__ as all_import
from .standard import __all__ as all_standard

__all__ = all_util+all_gate+all_types+all_import+all_standard

from .import_ket import code_ket



from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ... import *
from ...preprocessor import _ket_if, _ket_next
from math import sqrt, asin
from typing import Optional, Callable

def bell(x : int = 0, y : int = 0, qubits : Optional[quant] = None) -> quant:
    r"""Bell state preparation
    
    Return two entangle qubits in the Bell state:
    
    .. math::

        \left|\beta_{x,y}\right> = \frac{\left|0,y\right> + (-1)^x\left|1,¬y\right>}{\sqrt{2}}

    Args: 
        x: :math:`x`.
        y: :math:`y`.
        qubits: if provided, prepare the state in ``qubits``.
    """

    if qubits is not None and (len(qubits) != 2 or type(qubits) != quant):
        raise AttributeError("if param 'qubits' is provided, it must be a quant of length 2")

    if qubits is None:
        qubits = quant(2)

    if isinstance(x, future):
        end = _ket_if(x)
        X(qubits[0])
        _ket_next(end)
    elif x:
        X(qubits[0])

    if isinstance(y, future):
        end = _ket_if(y)
        X(qubits[1])
        _ket_next(end)
    elif y:
        X(qubits[1])

    H(qubits[0])
    ctrl(qubits[0], X, qubits[1])
    return qubits

def ghz(qubits : quant | int) -> quant:
    """GHZ state preparation"""
    if not isinstance(qubits, quant):
        qubits = quant(qubits)
    
    ctrl(H(qubits[0]), X, qubits[1:])

    return qubits

def w(qubits : quant | int) -> quant:
    """W state preparation"""
    if not isinstance(qubits, quant):
        size = qubits
        qubits = quant(qubits)
    size = len(qubits) 

    X(qubits[0])
    for i in range(size-1):
        with control(qubits[i]):
            n = size - i
            RY(2*asin(sqrt((n-1)/n)), qubits[i+1])
        cnot(qubits[i+1], qubits[i])

    return qubits

def pauli(basis : Callable, q : quant, state : int = +1) -> quant:
    """Prepare in the +1 or -1 eigenstate of a given Pauli operator.
    
    Args:
        basis: Pauli operator :func:`~ket.gates.X`, :func:`~ket.gates.Y`, or :func:`~ket.gates.Z`.
        q: Input qubits.
        state: Eigenstate.
    """

    if isinstance(state, future):
        end = _ket_if(state == -1)
        X(q)
        _ket_next(end)
    elif state == -1:
        X(q)
    elif state == 1:
        pass
    else:
        raise ValueError("Param 'state' must be +1 or -1.")

    if basis == X:
        return H(q)
    elif basis == Y:
        return S(H(q))
    elif basis == Z:
        return q
    else:
        raise ValueError("Param 'basis' must be X, Y, or Z.")








from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..ket import ctrl_begin, ctrl_end, X
from ..types import quant
from typing import Iterable, Callable, Optional, Any
from functools import reduce
from operator import add

def _create_mask(on_state : Optional[int | Iterable[int]], length : int) -> list[int]:
    """Create a mask for ctrl and control"""

    if on_state is None:
        return None
    elif hasattr(on_state, '__iter__'):
        if len(on_state) != length:
            raise ValueError(f"'on_state' received a list of length {len(on_state)} to use on {length} qubits")
        return on_state
    else:
        if length < on_state.bit_length():
            raise ValueError(f"To control 'on_state' {on_state} you need at least {on_state.bit_length()} qubits")
        return [int(i) for i in f"{{:0{length}b}}".format(on_state)]

def _apply_mask(mask : Iterable[int], q : quant):
    """Flip qubit q[i] if mask[i] == 0"""
    
    if mask is not None:
        for i, q in zip(mask, q):
            if i == 0:
                X(q)

class control:
    r"""Open a controlled-scope
    
    Inside a ``with control`` scope, the qubits of ``ctr`` control every
    quantum operation, only execution if all qubits are in the estate
    :math:`\left|1\right>`.

    Optionally, you can change the state when applying the quantum operation
    with the Keyword Argument ``on_state``.  If ``on_state`` is an ``int``, the
    number of qubits needs to be equal to or greater than the bit length of
    ``on_state``.  Else if ``on_sate`` is a ``List[int]``, its length needs to
    be equal to the number of qubits.

    For example, execute when the qubits are in the state
    :math:`\left|\left[\dots0\right]11\right>`, use ``with control(ctr, on_state=3):``
    or ``with control(ctr, on_state=[0, 1, 1]):`` if ``ctr`` has exactly 3 qubits.
    
    :Usage:

    .. code-block:: ket
    
        with control(*ctr[, on_state]):
            ...

    :Example:

    .. code-block:: ket

        c = quant(2)
        a, b = quant(2)

        # CNOT c[0] a
        with control(c[0]):
            X(a)
        
        # Toffoli c[0] c[1] a
        with control(c):
            X(a)

        # CSWAP c[0] a b
        with control(c[0]):
            swap(a, b)

    Args:
        ctr: Control qubits.
        on_state: Change the control state. 
    """

    def __init__(self, *ctr : quant, on_state : Optional[int | Iterable[int]] = None):
        self.ctr = reduce(add, ctr)
        self.mask = _create_mask(on_state, len(self.ctr))

    def __enter__ (self):
        _apply_mask(self.mask, self.ctr)
        ctrl_begin(self.ctr)
     
    def __exit__ (self, type, value, tb):
        ctrl_end()
        _apply_mask(self.mask, self.ctr)
            
def _ctrl(control  : quant | Iterable[quant], 
          func     : Callable | Iterable[Callable],
          *args, 
          on_state : Optional[int | list[int]] = None,
          **kwargs) -> Any:
    """Call Callable with controll-qubits"""

    ret = []
    mask = _create_mask(on_state, len(control))

    _apply_mask(mask, control)
    ctrl_begin(control)
    
    if hasattr(func, '__iter__'): 
        for f in func:
            ret.append(f(*args, **kwargs))
    else:
        ret = func(*args, **kwargs)

    ctrl_end()
    _apply_mask(mask, control)
    
    return ret

def _qubit_for_ctrl(qubits : quant | Iterable[quant] | slice | int | Iterable[int]) -> tuple[Callable[[quant], quant] | quant, bool]:
    """Get qubits for ctrl"""

    if any(isinstance(qubits, tp) for tp in [slice, int]):
        return lambda q : q[qubits], True
    elif hasattr(qubits, '__iter__') and all(isinstance(i, int) for i in qubits):
        return lambda q : q.at(qubits), True
    elif hasattr(qubits, '__iter__') and all(isinstance(i, quant) for i in qubits):
        return reduce(add, qubits), False
    else:
        return qubits, False

def ctrl(control    : quant | Iterable[quant] | slice | int | Iterable[int], 
         func       : Callable | Iterable[Callable] , 
         *args,    
         on_state   : Optional[int | Iterable[int]] = None,
         target     : Optional[slice | int | Iterable[int]] = None,
         later_call : bool = False,
         **kwargs) -> Callable | Any :
    r"""Add controll-qubits to a Callable

    :Call with control qubits:

    * ``control`` type must be :class:`~ket.types.quant` or ``Iterable[quant]``
    * ``func`` type must be ``Callable`` or ``Iterable[Callable]``

    .. code-block:: ket

        ret1 = ctrl(control_qubits, func, *args, **kwargs)
        # Equivalent to:
        # with control(control_qubits):
        #     ret1 = func(*args, **kwargs)

        ret2 = ctrl([q0, q1, q2, ...], func, *args, **kwargs)
        # Equivalent to:
        # with control(*control):
        #     ret2 = func(*args, **kwargs)
        
        ret3 = ctrl(control_qubits, [f0, f1, f2, ...], *args, **kwargs)
        # Equivalent to:
        # ret3 = []
        # with control(control_qubits):
        #     for f in func:
        #         ret3.append(f(*args, **kwargs))
 
    :Create controlled-operation:

    1. If the keyword argument ``later_call`` is ``True``, return a
    ``Callable[[], Any]``:

    .. code-block:: ket

        ctrl_func = ctrl(control_qubits, func, *args, **kwargs, later_call=True)
        # Equivalent to:
        # ctrl_func = lambda : ctrl(control_qubits, func, *args, **kwargs)

    Example:

    .. code-block:: ket
        :emphasize-lines: 8

        def increment(q):
            if len(q) > 1:
                ctrl(q[-1], increment, q[:-1])
            X(q[-1])

        size = ceil(log2(len(inputs)))+1
        with quant(size) as inc:
            with around(ctrl(q , increment, inc, later_call=True) for q in inputs):
                with control(inc, on_state=len(inputs)//2):
                    X(output)
            inc.free()

    2. If ``control`` and ``target`` type is ``int``, ``slice``, or
    ``Iterable[int]``,  return a ``Callable[[quant], Any]``:

    .. code-block:: ket

        ctrl_func = ctrl(ctrl_index, func, target_index)
        # Equivalent to:
        # ctrl_func = lambda q : ctrl(q[ctrl_index], func, q[target_index])

    Example:

    .. code-block:: ket

        with around(ctrl(0, X, slice(1, None)), q): # ctrl(q[0], X, q[1:]) 
            H(q[0])                                 # H(q[0])
                                                    # ctrl(q[0], X, q[1:])
 
    Args:
        control: Control qubits.
        func: Functions to add control.
        on_state: Change the control state, same as for :class:`~ket.standard.control`. 
        target: Target qubits to create controlled-operations.
        later_call: If ``True``, do not execute and return a ``Callable[[], Any]``.
    """

    control, create_gate = _qubit_for_ctrl(control)
    if create_gate:
        target, _ = _qubit_for_ctrl(*args if target is None else [target])
        def _ctrl_gate(q : quant) -> quant:
            _ctrl(control(q), func, target(q), on_state=on_state)
            return q
        return _ctrl_gate
    elif later_call:
        return lambda : _ctrl(control, func, *args, on_state=on_state, **kwargs)
    else:
        return _ctrl(control, func, *args, on_state=on_state, **kwargs)



from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..ket import plugin
from ..types import quant
from inspect import getsource
from base64 import b64encode
from textwrap import dedent
from typing import Optional

__all__ = ['plugin', 'pown', 'diagonal', 'make_quantum']

def pown(a : int, x : quant, N : int) -> quant: 
    r"""Apply a modular exponentiation in a superposition.

    .. math::

        \left| x \right> \left| 0 \right> \rightarrow  \left| x \right> \left| a^x\; \text{mod} \, N \right>

    Args:
        a: :math:`a`.
        x: :math:`x`.
        N: :math:`N`.

    :return: :class:`~ket.types.quant` with the operation result.
    """
    
    ret = quant(N.bit_length())
    arg = str(len(ret)) + ' ' + str(a) + ' ' + str(N)

    plugin('ket_pown', x|ret, arg)

    return ret

def diagonal(diag : list[float], q : quant):
    r"""Apply a diagonal matrix.

    .. math::

        \begin{bmatrix}
                 e^{i\lambda_0} & 0              & \cdots & 0 \\
                 0              & e^{i\lambda_1} & \cdots & 0 \\
                 \vdots         & \vdots         & \ddots & \vdots \\
                 0              & 0              & \cdots & e^{i\lambda_{2^n-1}} 
             \end{bmatrix}

    Args:
        diag: :math:`\left[ \lambda_0, \lambda_1, \dots, \lambda_{2^n-1} \right]`
        q: Input qubits.
    """
    
    arg = ' '.join(str(i) for i in diag)
    plugin('ket_diag', q, arg)
    
def matrix(u00 : complex, u01 : complex, u10 : complex, u11 :complex, q : Optional[quant] = None):
    mat = [u00, u01, u10, u11]
    arg = ' '.join(str(i.real) for i in mat) +' '+' '.join(str(i.imag) for i in mat)
    if q is None:
        return lambda q : plugin('ket_mat', q, arg)
    else:
        plugin('ket_mat', q, arg)

def make_quantum(*args, **kwargs):
    r'''Make a Python function operate with quant variables.

    :Usage:

    .. code-block:: ket

        @make_quantum 
        def pow7mod15(reg1, reg2):
            reg2 = pow(7, reg1, 15)
            return reg1, reg2

        reg1, reg2 = quant(4), quant(4)
        pow7mod15(reg1, reg2)
    
    .. code-block:: ket

        balanced_oracle = make_quantum(func="""
            def oracle(a, b):
                def f(a):
                    return a
                return a, f(a)^b
        """, name='oracle')
        a, b = quant(2)
        balanced_oracle(a, b)
    
    '''
    
    if len(args) != 0 and callable(args[0]):
        func_str = '\n'.join(getsource(args[0]).split('\n')[1:])
        func_name = args[0].__name__
    else:
        func_str = kwargs['func']
        func_name = kwargs['name']

    func_b64 = b64encode(dedent(func_str).encode()).decode()
    def inner(*args):
        plugin_args = str(len(args)) + ' '
        for q in args:
            plugin_args += str(len(q)) + ' '
        plugin_args += func_name + ' ' + func_b64

        qubits = args[0]
        for q in args[1:]:
            qubits |= q
        plugin('ket_pycall', qubits, plugin_args)
    
    return inner




#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..ket import quant as _quant, measure, quant_dirty
from typing import Iterable
from functools import reduce
from operator import add

class quant(_quant):
    r"""Qubit list

    Allocate ``size`` qubits in the state :math:`\left|0\right>` and return
    its reference in a new :class:`~ket.types.quant`.

    Qubits allocated using a ``with`` statement must be released at the end of
    the scope.
    
    Example:
    
    .. code-block:: ket

        a = H(quant()) 
        b = X(quant())
        with quant() as aux: 
            with around(H, aux):
                with control(aux):
                    swap(a, b)
            result = measure(aux)
            if result == 1:
                X(aux) 
            aux.free() 

    :Qubit Indexing:

    Use brackets to index qubits as in a ``list`` and use ``+`` to concatenate
    two :class:`~ket.types.quant`.

    Example:

    .. code-block:: ket

        q = quant(20)        
        head, tail = q[0], q[1:]
        init, last = q[:-1], q[-1]
        even = q[::2]
        odd = q[1::2]
        reverse = reversed(q) # invert qubits order

        
        a, b = quant(2) # |a⟩, |b⟩
        c = a+b         # |c⟩ = |a⟩|b⟩ 

    Args:
        size: The number of qubits to allocate.
    """

    def __init__(self, size : int = 1):
        r"""Allocate `size` qubits"""
        super().__init__(size)
        
    @staticmethod
    def dirty(size : int = 1):
        """Allocate dirty qubits

        Allocate ``size`` qubits at an unknown state. 

        warning:
            Use dirty qubits may have side effects due to previous entanglement.

        Args:
            size: The number of dirty qubits to allocate.
        """
        ret = quant_dirty(size)
        ret.__class__ = quant
        return ret

    def __add__(self, other):
        ret = super().__or__(other)
        ret.__class__ = quant
        return ret

    def __or__(self, other):
        ret = super().__or__(other)
        ret.__class__ = quant
        return ret

    def at(self, index : Iterable[int]):
        r"""Return qubits at `index`
        
        Create a new :class:`~ket.types.quant` with the qubit references at the
        position defined by the ``index`` list.

        :Example:

        .. code-block:: ket

            q = quant(20)        
            odd = q.at(range(1, len(q), 2)) # = q[1::2]

        Args:
            index: List of indexes.
        """

        return reduce(add, [self[i] for i in index])   
    
    def free(self, dirty : bool = False):
        r"""Free the qubits

        All qubits must be at the state :math:`\left|0\right>` before the call,
        otherwise set the ``dirty`` param to ``True``.

        Warning: 
            No check is applied to see if the qubits are at state
            :math:`\left|0\right>`.

        Args:
            dirty: Set ``True`` to free dirty qubits.
        """
 
        super().free(dirty)
        
    def is_free(self) -> bool:
        """Return ``True`` when all qubits are free"""
        return super().is_free()

    def __reversed__(self):
        ret = super().inverted()
        ret.__class__ = quant
        return ret

    def __getitem__(self, key):
        ret = super().__getitem__(key)
        ret.__class__ = quant
        return ret

    class quant_iter:
        def __init__(self, q):
            self.q = q
            self.idx = -1
            self.size = len(q)

        def __next__(self): 
            self.idx += 1
            if self.idx < self.size:
                return self.q[self.idx]
            raise StopIteration

        def __iter__(self):
            return self
 
    def __iter__(self):
        return self.quant_iter(self)

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        if not self.is_free():
            raise RuntimeError('non-free quant at the end of scope')

    def __int__(self):
        return measure(self).get()

    def __repr__(self):
        return f"<Ket type 'quant'; {str(len(self))} qubits; {super().__repr__()}>"



from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .. import *
from .. import code_ket
from math import pi, sqrt
from typing import Callable, Iterable

def qft(q : quant, invert : bool = True) -> quant:
    """Quantum Fourier Transformation
    
    Apply a QFT_ on the qubits of q.

    Note:
        Implace operation.

    .. _QFT: https://en.wikipedia.org/wiki/Quantum_Fourier_transform

    :param q: input qubits
    :param invert: if ``True``, invert qubits with swap gates
    :return: qubits in the reserved order 
    """

    if len(q) == 1:
        H(q)
    else:
        head, *tail = q
        H(head)
        for i in range(len(tail)):
            with control(tail[i]):
                phase(2*pi/2**(i+2), head)
        qft(tail, invert=False)

    if invert:
        for i in range(len(q)//2):
            swap(q[i], q[len(q)-i-1])
        return q
    else:
        return reversed(q)

def dump_matrix(u : Callable | Iterable[Callable], size : int = 1) -> list[list[complex]]:
    """Get the matrix of a quantum operation.
    
    Args:
        u: Quantum operation.
        size: Number of qubits.
    """

    mat = [[0 for _ in range(2**size)] for _ in range(2**size)]
    with run():
        row = quant(size)
        column = quant(size)
        H(column)
        cnot(column, row)
        if hasattr(u, '__iter__'):
            ret = []
            for ui in u:
               ret.append(ui(row)) 
        else:
            ret = u(row)
        d = dump(column|row)
        exec_quantum()
    
    for state in d.get_states():
        column = state >> size
        row = state & ((1 << size)-1)
        mat[row][column] = d.amplitude(state)[0]/(1/sqrt(2**size))

    if ret != None:
        return mat, ret
    else:
        return mat
     


from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..ket import X as _X, Y as _Y, Z as _Z, H as _H, S as _S, SD as _SD, T as _T, TD as _TD
from ..ket import phase as _phase, RX as _RX, RY as _RY, RZ as _RZ 
from ..types import quant
from ..standard import ctrl, around
from typing import Callable, Optional, Iterable
from functools import reduce
from operator import add
from math import pi

__all__ = ['I', 'X', 'Y', 'Z', 'H', 'S', 'SD', 'T', 'TD', 'phase', 'RX', 'RY', 'RZ', 'cnot', 'swap', 'RXX', 'RYY', 'RZZ']

def X(q : quant | Iterable[quant]) -> quant:
    r""" Pauli-X gate (:math:`\sigma_x`)

    Apply a single-qubit Pauli-X gate on every qubit of ``q``.

    :Matrix Representation:

    .. math::

        X = \begin{bmatrix} 
                0 & 1 \\
                1 & 0 
            \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix} 
            X\left|0\right> = & \left|1\right> \\
            X\left|1\right> = & \left|0\right>
        \end{matrix}

    :param q: Input qubits.
    """
    q = reduce(add, q)
    _X(q)
    return q

def Y(q : quant | Iterable[quant]) -> quant:
    r"""Pauli-Y gate (:math:`\sigma_y`)

    Apply a single-qubit Pauli-Y gate on every qubit of ``q``.

    :Matrix Representation:

    .. math::

        Y = \begin{bmatrix} 
                   0 & -i \\
                   i & 0 
            \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix} 
            Y\left|0\right> = & -i\left|1\right> \\
            Y\left|1\right> =& i\left|1\right>
        \end{matrix}

    :param q: Input qubits.
    """
    q = reduce(add, q)
    _Y(q)
    return q

def Z(q : quant | Iterable[quant]) -> quant:
    r"""Pauli-Z gate (:math:`\sigma_z`)

    Apply a single-qubit Pauli-Y gate on every qubit of ``q``.

    :Matrix Representation:

    .. math::

        Z = \begin{bmatrix} 
                1 & 0 \\
                0 & -1 
            \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix} 
            Z\left|0\right> = & \left|0\right> \\ 
            Z\left|1\right> = & -\left|1\right>
        \end{matrix}

    :param q: Input qubits.
    """
    q = reduce(add, q)
    _Z(q)
    return q

def H(q : quant | Iterable[quant]) -> quant:
    r"""Hadamard gate

    Apply a single-qubit Hadamard gate on every qubit of ``q``.

    :Matrix Representation:

    .. math::

        H = \frac{1}{\sqrt{2}}\begin{bmatrix} 
                                    1 & 1 \\
                                    1 & -1 
                              \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix} 
            H\left|0\right> = & \frac{\left|0\right>+\left|1\right>}{\sqrt{2}} = & \left|+\right> \\
            H\left|1\right> = & \frac{\left|0\right>-\left|1\right>}{\sqrt{2}} = & \left|-\right> \\
            H\left|+\right> = & \left|0\right> \\
            H\left|-\right> = & \left|1\right> \\
        \end{matrix}

    :param q: Input qubits.
    """
    q = reduce(add, q)
    _H(q)
    return q

def S(q : quant | Iterable[quant]) -> quant:
    r"""S gate 

    Apply a single-qubit S gate on every qubit of ``q``.

    :Matrix Representation:

    .. math::

        S = \begin{bmatrix} 
                1 & 0 \\
                0 & i 
            \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix} 
            S\left|0\right> = & \left|0\right> \\
            S\left|1\right> = & i\left|1\right> 
        \end{matrix}

    :param q: Input qubits.
    """
    q = reduce(add, q)
    _S(q)
    return q

def SD(q : quant | Iterable[quant]) -> quant:
    r"""S-dagger gate (:math:`S^\dagger`)

    Apply a single-qubit S-dagger gate on every qubit of ``q``.
    
    ``SD`` :math:`\equiv` ``adj(S)``

    :Matrix Representation:

    .. math::

        S^\dagger = \begin{bmatrix} 
                        1 & 0 \\
                        0 & -i 
                    \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix} 
                 S^\dagger\left|0\right> = & \left|0\right> \\
                 S^\dagger\left|1\right> = & -i\left|1\right> 
             \end{matrix}

    :param q: Input qubits.
    """
    q = reduce(add, q)
    _SD(q)
    return q

def T(q : quant | Iterable[quant]) -> quant:
    r"""T gate

    Apply a single-qubit T gate on every qubit of ``q``.

    :Matrix Representation:

    .. math::

         T = \begin{bmatrix} 
                    1 & 0 \\
                    0 & e^{i\pi/4} 
             \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix} 
            T\left|0\right> = & \left|0\right> \\
            T\left|1\right> = & \frac{1+i}{\sqrt{2}}\left|1\right> \end{matrix}

    :param q: Input qubits.
    """
    q = reduce(add, q)
    _T(q)
    return q

def TD(q : quant | Iterable[quant]) -> quant:
    r"""T-dagger gate (:math:`T^\dagger`)

    Apply a single-qubit T-dagger gate on every qubit of ``q``.

    ``ST`` :math:`\equiv` ``adj(T)``

    :Matrix Representation:

    .. math::

         T^\dagger = \begin{bmatrix} 
                        1 & 0 \\
                        0 & e^{-i\pi/4}
                     \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix} 
                T^\dagger\left|0\right> = & \left|0\right> \\
                T^\dagger\left|1\right> = & \frac{1-i}{\sqrt{2}}\left|1\right> \end{matrix}

    :param q: Input qubits.
    """
    q = reduce(add, q)
    _TD(q)
    return q

def phase(lamb : float, q : Optional[quant | Iterable[quant]] = None) -> Callable[[quant], quant] | quant:
    r"""Phase gate

    Apply a relative phase of :math:`e^{i\lambda}` on every qubit of ``q``.

    If ``q`` is not provided, return a new gate: ``phase(lamb)(q)``
    :math:`\equiv` ``phase(lamb, q)``. 


    :Matrix Representation:

    .. math::

        P(\lambda) = \begin{bmatrix}
                        1 & 0 \\
                        0 & e^{i\lambda}
                     \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix}
                P\left|0\right> = & \left|0\right> \\
                P\left|1\right> =& e^{i\lambda}\left|1\right> 
        \end{matrix}

    :param q: Input qubits.
    :param lamb: Relative phase :math:`\lambda`.
    """
    def __phase(q):
        q = reduce(add, q)
        _phase(lamb, q)
        return q
      
    return __phase(q) if q is not None else __phase

def RX(theta : float, q : Optional[quant | Iterable[quant]] = None) -> Callable[[quant], quant] | quant:
    r"""X-axis rotation gate

    Apply a rotation of :math:`\theta` about the X-axis on every qubit of ``q``.

    If ``q`` is not provided, return a new gate: ``RX(theta)(q)``
    :math:`\equiv` ``RX(theta, q)``. 

    :Matrix Representation:

    .. math::

         RX(\theta) = \begin{bmatrix} 
                        \cos{\frac{\theta}{2}} & -i\sin{\frac{\theta}{2}} \\
                        -i\sin{\frac{\theta}{2}} & \cos{\frac{\theta}{2}} 
                      \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix}
                RX\left|0\right> = & \cos\frac{\theta}{2}\left|0\right> -i\sin\frac{\theta}{2}\left|1\right> \\
                RX\left|1\right> =& -i\sin\frac{\theta}{2}\left|0\right> + \cos\frac{\theta}{2}\left|1\right>
        \end{matrix}

    :param q: Input qubits.
    :param theta: Rotation angle :math:`\theta`.
    """
    def __RX(q):
        q = reduce(add, q)
        _RX(theta, q)
        return q
     
    return __RX(q) if q is not None else __RX

def RY(theta : float, q : Optional[quant | Iterable[quant]] = None) -> Callable[[quant], quant] | quant:
    r"""Y-axis rotation gate

    Apply a rotation of :math:`\theta` about the Y-axis on every qubit of ``q``.

    If ``q`` is not provided, return a new gate: ``RY(theta)(q)``
    :math:`\equiv` ``RY(theta, q)``. 

    :Matrix Representation:

    .. math::

         RY(\theta) = \begin{bmatrix} 
                        \cos{\frac{\theta}{2}} & -\sin{\frac{\theta}{2}} \\
                        \sin{\frac{\theta}{2}} & \cos{\frac{\theta}{2}} 
                      \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix}
            RY\left|0\right> = & \cos{\theta\over2}\left|0\right> -i\sin\frac{\theta}{2}\left|1\right> \\
            RY\left|1\right> =& -\sin\frac{\theta}{2}\left|0\right> + \cos\frac{\theta}{2}\left|1\right> 
        \end{matrix}

    :param q: Input qubits.
    :param theta: Rotation angle :math:`\theta`.
    """
    def __RY(q):
        q = reduce(add, q)
        _RY(theta, q)
        return q
      
    return __RY(q) if q is not None else __RY

def RZ(theta : float, q : Optional[quant | Iterable[quant]] = None) -> Callable[[quant], quant] | quant:
    r"""Z-axis rotation gate

    Apply a rotation of :math:`\theta` about the Z-axis on every qubit of ``q``.

    If ``q`` is not provided, return a new gate: ``RZ(theta)(q)``
    :math:`\equiv` ``RZ(theta, q)``. 

    :Matrix Representation:

    .. math::

         RZ(\theta) =  \begin{bmatrix} 
                            e^{-i\theta/2} & 0 \\
                            0 & e^{i\theta/2} 
                       \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix}
            RZ\left|0\right> = & e^{-i\theta/2}\left|0\right> \\ 
            RZ\left|1\right> =& e^{i\theta/2}\left|1\right> 
        \end{matrix}

    :param q: Input qubits.
    :param theta: Rotation angle :math:`\theta`.
    """
    def __RZ(q):
        q = reduce(add, q)
        _RZ(theta, q)
        return q
    
    return __RZ(q) if q is not None else __RZ


def I(q : quant | Iterable[quant]) -> quant:
    r"""Identity gate 

    Apply an identity gate on every qubit of ``q``.

    Note:
        This quantum gate does nothing! 

    :Matrix Representation:

    .. math::

        I = \begin{bmatrix} 
                1 & 0 \\
                0 & 1 
            \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix} 
            I\left|0\right> = & \left|0\right> \\
            I\left|1\right> = & \left|1\right>
        \end{matrix}

    :param q: Input qubits.
    """
    q = reduce(add, q)
    return q
 
def cnot(c : quant | Iterable[quant], t : quant | Iterable[quant]) -> tuple[quant, quant]:
    r"""Controlled-NOT
    
    Apply a CNOT gate between the qubits of c and t.

    :Matrix Representation:

    .. math::

        CNOT = \begin{bmatrix} 
                    1 & 0 & 0 & 0\\
                    0 & 1 & 0 & 0\\
                    0 & 0 & 0 & 1\\
                    0 & 0 & 1 & 0
                \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix} 
            CNOT\left|c t\right> = & \left|c (c\oplus t)\right> 
        \end{matrix}

    :param c: Control qubits.
    :param t: Target qubits.
    """
    
    c = reduce(add, c)
    t = reduce(add, t)
    
    for i, j in zip(c, t):
        ctrl(i, X, j)
    
    return c, t

def swap(a : quant | Iterable[quant], b : quant | Iterable[quant]) -> tuple[quant, quant]:
    r"""SWAP gate
    
    Swap the qubits of a and b.

    :Matrix Representation:

    .. math::

        CNOT = \begin{bmatrix} 
                    1 & 0 & 0 & 0\\
                    0 & 0 & 1 & 0\\
                    0 & 1 & 0 & 0\\
                    0 & 0 & 0 & 1
                \end{bmatrix}

    :Effect:

    .. math::

        \begin{matrix} 
            CNOT\left|a b\right> = & \left|b a\right> 
        \end{matrix}

    :param a: Input qubits.
    :param b: Input qubits.
    """
    a = reduce(add, a)
    b = reduce(add, b)

    cnot(a, b)
    cnot(b, a)
    cnot(a, b)

    return a, b

def RXX(theta : float, a : quant | Iterable[quant], b : quant | Iterable[quant]) -> tuple[quant, quant]:
    r"""XX-axis rotation gate
    

    :Matrix Representation:

    .. math::

        RXX(\theta) = \begin{bmatrix}
                          \cos\theta   & 0            & 0            & -i\sin\theta \\
                          0            & \cos\theta   & -i\sin\theta & 0 \\
                          0            & -i\sin\theta & \cos\theta   & 0 \\
                          -i\sin\theta & 0            & 0            & \cos\theta
                      \end{bmatrix}
  
    :param theta: :math:`\theta`.
    :param a: Input qubits.
    :param b: Input qubits.
    """
    a = reduce(add, a)
    b = reduce(add, b)
    
    for qa, qb in zip(a, b):
        with around([H, ctrl(0, X, 1)], qa+qb):
            RZ(theta, qb)

    return a, b 

def RYY(theta : float, a : quant | Iterable[quant], b : quant | Iterable[quant]) -> tuple[quant, quant]:
    r"""YY-axis rotation gate
    
    :param theta: :math:`\theta`.
    :param a: Input qubits.
    :param b: Input qubits.
    """
    a = reduce(add, a)
    b = reduce(add, b)
    
    for qa, qb in zip(a, b):
        with around([RX(pi/2), ctrl(0, X, 1)], qa+qb):
            RZ(theta, qb)

    return a, b 

def RZZ(theta : float, a : quant | Iterable[quant], b : quant | Iterable[quant]) -> tuple[quant, quant]:
    r"""ZZ-axis rotation gate
    
    :param theta: :math:`\theta`.
    :param a: Input qubits.
    :param b: Input qubits.
    """
    a = reduce(add, a)
    b = reduce(add, b)
    
    for qa, qb in zip(a, b):
        with around(ctrl(0, X, 1), qa+qb):
            RZ(theta, qb)
            
    return a, b 


import subprocess
import setuptools
import sys
from datetime import datetime

def git_hash():
    try:
        return subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).decode()[:-1] + ", "
    except:
        return ""

try:
    from skbuild import setup
except ImportError:
    subprocess.call([sys.executable, '-m', 'pip', 'install', 'scikit-build'])
    from skbuild import setup

with open('README.md', 'r') as file:
    long_description = file.read()

setup_requirements = [
    'scikit-build>=0.11',
    'cmake>=3.15', 
    'ninja>=1.10', 
    'conan>=1.25'
]

__version__ = '0.1'

setup(
    name            = 'ket-lang',
    description     = 'Ket Quantum Programming Language interpreter and library.',
    url             = 'https://quantum-ket.gitlab.io',
    version         = __version__,
    author          = 'Evandro Chagas Ribeiro da Rosa',
    author_email    = 'evandro.crr@posgrad.ufsc.br',
    license         = 'Apache-2.0',
    cmake_source_dir='.',
    cmake_args=[
        '-DCMAKE_BUILD_TYPE=Release',
        f'-DKET_BUILD_INFO="{__version__} ({git_hash()}{datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")})"',
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_namespace_packages(include=['ket', 'ket.*']),
    setup_requires=setup_requirements,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: C++',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
    ],
    entry_points={'console_scripts': ['ket = ket.__main__:__ket__']},
)



#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .import_ket import _import_globals_ket
from . import *
from .ket import label, branch, jump, build_info
from .preprocessor import *

def __ket__():
    import argparse
    from os import path, getcwd

    parser = argparse.ArgumentParser(prog='ket', description='Ket interprester')
    parser.add_argument('--version', action='version', version=f'Ket {build_info()}')
    parser.add_argument('input', metavar='.ket', nargs=argparse.REMAINDER, type=str, help='source code')

    args = parser.parse_args()

    if len(args.input) == 0:
        print("No input")
        exit(1)

    globals()['__name__'] = '__main__'
    globals()['__in_ket__'] = True
    source = path.join(getcwd(), args.input[0])
    _import_globals_ket(source, globals())

if __name__ == '__main__':
    __ket__()
